@@ -21,7 +21,7 @@ and facilitate development:
 """
 
 # Local imports
-from . import database, socket_comm, __project__
+from . import database, socket_comm, custom_logging, __project__
 
 __all__ = [
     __project__.__author__,
@@ -31,4 +31,5 @@ __all__ = [
     __project__.__project_name__,
     'database',
     'socket_comm',
+    'custom_logging',
 ]
