@@ -9,13 +9,24 @@ The database settings are set with a
 and the standard library :obj:`configparser`.
 """
 
+# Imports
 import configparser
-import logging
-from pkg_resources import resource_filename
+from os.path import abspath, expanduser
 from enum import Enum
 
-import psycopg2
-from psycopg2 import sql
+# Third party
+from psycopg2 import (
+    sql,
+    Error as PSQL_Error,
+    DatabaseError,
+    IntegrityError,
+    OperationalError,
+    DataError,
+    connect,
+)
+
+# Local packages
+from lab_utils.custom_logging import getLogger
 
 
 class DataType(Enum):
@@ -36,7 +47,8 @@ class DataType(Enum):
     time = 'TIMESTAMPTZ'    #: Time stamp, with time zone information
 
     def __str__(self) -> str:
-        """ Parse the data type as an SQL string
+        """ Parses the data type as an SQL string.
+
         Returns
         -------
         str
@@ -54,7 +66,8 @@ class Constraint(Enum):
     positive_strict = ''' CHECK({column_name} > 0) '''      #: The variable must be strictly positive
 
     def __str__(self) -> str:
-        """ Parse the constraint as an SQL string
+        """ Parses the constraint as an SQL string.
+
         Returns
         -------
         str
@@ -65,20 +78,20 @@ class Constraint(Enum):
 
 class Database:
     """Manages connections and operations with a PostgreSQL database.
-    The class is based on the library :obj:`psycopg2` and
+    The class is based on the :obj:`psycopg2` library and
     on this `tutorial
     <https://www.postgresqltutorial.com/postgresql-python/>`_.
     """
 
-    _host: str = 'localhost'        #: The host name where the database is located
-    _port: int = 5432               #: Connection port
-    _database: str = 'beam'         #: The database name to connect to
-    _user: str = 'cw-beam'          #: User name
-    _passfile: str = '~/.pgpass'    #: Location of the pgpass file with the credentials
+    host: str = 'localhost'         #: The host name where the database is located
+    port: int = 5432                #: Connection port
+    database: str = 'beam'          #: The database name to connect to
+    user: str = 'cw-beam'           #: User name
+    passfile: str = '~/.pgpass'     #: Location of the pgpass file with the credentials
 
-    _connection = None              #: Connection object returned by psycopg2.connect()
-    _db_version: str = ''           #: Database version
-    _cursor = None                  #: Cursor provided by _connection.cursor() to execute an SQL query
+    connection = None               #: Connection object returned by psycopg2.connect()
+    db_version: str = ''            #: Database version
+    cursor = None                   #: Cursor provided by _connection.cursor() to execute an SQL query
 
     # List of predefined SQL queries
     _query_createTimeTable = '''
@@ -111,7 +124,8 @@ class Database:
         WHERE
         table_schema = %(table_schema)s
         AND
-        table_name = %(table_name)s)
+        table_name = %(table_name)s
+        )
         ;
     '''
     _query_addColumn = '''
@@ -127,7 +141,14 @@ class Database:
         ;
     '''
 
-    def __init__(self, config_file: str = None):
+    def __init__(self,
+                 config_file: str = None,
+                 host: str = None,
+                 port: int = None,
+                 database: str = None,
+                 user: str = None,
+                 passfile: str = None,
+                 ):
         """ Initializes the :class:`Database` object. If a
         :paramref:`configuration file name<Database.__init__.config_file>`
         is given, the constructor calls the method
@@ -135,10 +156,25 @@ class Database:
 
         Parameters
         ----------
-            config_file : str, optional
-                Configuration file name, default is `None`. See
-                :doc:`here<../../../examples/conf/conf_database>`
-                for a configuration file example
+        config_file : str, optional
+            Configuration file name, default is `None`. See
+            :doc:`here<../../../examples/conf/conf_database>`
+            for a configuration file example.
+
+        host : str, optional
+            Database host.
+
+        port : int, optional
+            Database port.
+
+        database : str, optional
+            Database name.
+
+        user : str, optional
+            Database user.
+
+        passfile : str, optional
+            Local file with PostgreSQL password.
 
         Raises
         ------
@@ -151,6 +187,22 @@ class Database:
         if config_file is not None:
             self.config(config_file)
 
+        # Override config file if other arguments are given
+        if host is not None:
+            self.host = host
+
+        if port is not None:
+            self.port = port
+
+        if database is not None:
+            self.database = database
+
+        if user is not None:
+            self.user = user
+
+        if passfile is not None:
+            self.passfile = passfile
+
     def __del__(self):
         """ Closes the connection to the database, if it was ever open.
 
@@ -161,19 +213,18 @@ class Database:
         """
         self.close()
 
-    def config(self, config_file: str = resource_filename(__name__, 'conf/database.ini')):
+    def config(self, config_file: str):
         """ Loads the configuration.
 
-        The method reads the :paramref:`config_file`
-        using the library :obj:`configparser`. The
-        structure of the file is shown in the
+        Reads the :paramref:`config_file` using the
+        :obj:`configparser` library. The structure
+        of the file is shown in the
         :ref:`examples section<configuration-files>`.
 
         Parameters
         ----------
-        config_file : str, optional
-            :doc:`Configuration file name<../../../examples/conf/conf_database>`,
-            default is 'conf/database.ini'
+        config_file : str
+            Configuration file name.
 
         Raises
         ------
@@ -182,31 +233,37 @@ class Database:
             a parameter is missing or it has an invalid value.
         """
 
-        # Use a logger named like the module itself
-        logger = logging.getLogger(__name__)
-        logger.info("Loading configuration file %s", config_file)
-
         try:
+            # Expand configuration file path
+            config_file = abspath(expanduser(config_file))
+            getLogger().info("Loading configuration file %s", config_file)
+
             # Initialize config parser and read file
             config_parser = configparser.ConfigParser()
             config_parser.read(config_file)
 
             # Assign values to class attributes
-            self._host = config_parser.get(section='Overall', option='host', fallback='localhost')
-            self._port = config_parser.getint(section='Overall', option='port', fallback=5432)
-            self._database = config_parser.get(section='Overall', option='database', fallback='beam')
-            self._user = config_parser.get(section='Overall', option='user', fallback='cw-beam')
-            self._passfile = config_parser.get(section='Overall', option='passfile', fallback='~/.pgpass')
+            self.host = config_parser.get(section='Overall', option='host')
+            self.port = config_parser.getint(section='Overall', option='port')
+            self.database = config_parser.get(section='Overall', option='database')
+            self.user = config_parser.get(section='Overall', option='user')
+            self.passfile = config_parser.get(section='Overall', option='passfile')
+
+            # Expand passfile path
+            self.passfile = abspath(expanduser(self.passfile))
 
         except configparser.Error as e:
-            logger.error("{}: {}".format(type(e).__name__, e))
+            getLogger().error("{}: {}".format(type(e).__name__, e))
             raise
 
         except BaseException as e:
             # Undefined exception, full traceback to be printed
-            logger.exception("{}: {}".format(type(e).__name__, e))
+            getLogger().exception("{}: {}".format(type(e).__name__, e))
             raise
 
+        else:
+            getLogger().info('Configuration loaded')
+
     def connect(self, print_version: bool = False):
         """ Connects to the database.
 
@@ -217,50 +274,51 @@ class Database:
         Parameters
         ----------
         print_version : bool, optional
-            Print the database version, default is False.
+            Print the database version, default is 'False'.
 
         Raises
         ------
         :class:`psycopg2.Error`
             Base exception for all kind of database errors
         """
-        # Use a logger named like the module itself
-        logger = logging.getLogger(__name__)
-        logger.info('Connecting to database <%s> on host <%s> over port <%d> as user <%s>',
-                    self._database,
-                    self._host,
-                    self._port,
-                    self._user)
+        getLogger().info(
+            'Connecting to database <%s> on host <%s> over port <%d> as user <%s>',
+            self.database,
+            self.host,
+            self.port,
+            self.user
+        )
 
         try:
             # Connect to the database
-            logger.debug(self._passfile)
-            self._connection = psycopg2.connect(
-                dbname=self._database,
-                user=self._user,
-                host=self._host,
-                port=self._port,
-                passfile=self._passfile)
+            self.connection = connect(
+                dbname=self.database,
+                user=self.user,
+                host=self.host,
+                port=self.port,
+                passfile=self.passfile
+            )
 
             # Read database version, serves as connection check
-            self._cursor = self._connection.cursor()
-            self._cursor.execute('SELECT version()')
-            self._db_version = self._cursor.fetchone()
+            self.cursor = self.connection.cursor()
+            self.cursor.execute('SELECT version()')
+            self.db_version = self.cursor.fetchone()
 
-        except psycopg2.Error as e:
-            logger.error("{}: {}".format(type(e).__name__, e))
+        except PSQL_Error as e:
+            getLogger().error("{}: {}".format(type(e).__name__, e))
             raise
 
         except BaseException as e:
             # Undefined exception, full traceback to be printed
-            logger.exception("{}: {}".format(type(e).__name__, e))
+            getLogger().exception("{}: {}".format(type(e).__name__, e))
             raise
 
         else:
             if print_version:
-                logger.info('Connection successful, database version %s', self._db_version)
+                getLogger().info('Connection successful, database version: %s', self.db_version)
             else:
-                logger.info('Connection successful')
+                getLogger().info('Connection successful')
+                getLogger().debug('Database version: %s', self.db_version)
 
     def close(self):
         """ Closes the connection to the database.
@@ -272,20 +330,21 @@ class Database:
         """
 
         # Close, if the connection had been opened
-        if self._connection is not None:
-            logger = logging.getLogger(__name__)
-            logger.info('Closing connection to database %s', self._database)
+        if self.connection is not None:
+            getLogger().info('Closing connection to database <%s>', self.database)
             try:
-                self._connection.close()
-            except psycopg2.Error as e:
-                logger.error("{}: {}".format(type(e).__name__, e))
+                self.connection.close()
+            except PSQL_Error as e:
+                getLogger().error("{}: {}".format(type(e).__name__, e))
                 raise
             except BaseException as e:
                 # Undefined exception, full traceback to be printed
-                logger.exception("{}: {}".format(type(e).__name__, e))
+                getLogger().exception("{}: {}".format(type(e).__name__, e))
                 raise
+            else:
+                getLogger().info('Connection to database <%s> closed', self.database)
             finally:
-                self._connection = None
+                self.connection = None
 
     def create_timescale_db(self, table_name: str, default_now: bool = True):
         """ Creates a
@@ -302,7 +361,7 @@ class Database:
         table_name : str
             The name of the table to be checked
         default_now : bool, optional
-            Set the 'time' column default to 'NOW()', default is True.
+            Set the 'time' column default to 'NOW()', default is `True`.
 
         Raises
         ------
@@ -310,40 +369,43 @@ class Database:
             Base exception for all kind of database errors
         """
 
-        logger = logging.getLogger(__name__)
+        logger = getLogger()
 
-        if self._connection is None:
-            raise psycopg2.DatabaseError('Trying to access database without initialization')
+        # Check database status
+        if self.connection is None:
+            raise DatabaseError('Trying to access database without initialization')
 
         try:
             # Prepare query for table creation
             if default_now:
-                logger.debug('Creating TimescaleDB table \'%s\' with default \'time\' NOW()', table_name)
+                logger.info('Creating TimescaleDB table <%s> with default <time> NOW()', table_name)
                 query = sql.SQL(self._query_createTimeTable.format(
                     table_name=table_name,
-                    default=sql.SQL('DEFAULT NOW()').as_string(self._connection)
+                    default=sql.SQL('DEFAULT NOW()').as_string(self.connection)
                 ))
             else:
-                logger.debug('Creating TimescaleDB table \'%s\' without default \'time\'', table_name)
+                logger.info('Creating TimescaleDB table <%s> without default <time>', table_name)
                 query = sql.SQL(self._query_createTimeTable.format(
                     table_name=table_name,
                     default=''
                 ))
 
             # Create the table table
-            self._cursor.execute(query)
+            logger.debug('Sending query: %s', query)
+            self.cursor.execute(query)
 
             # Make the table a TimescaleDB Hypertable
-            self._cursor.execute(
+            self.cursor.execute(
                 self._query_makeTimescaleHypertable,
                 {
                     'table_name':   table_name,
                     'key':          'time'
                 }
             )
-            self._connection.commit()
+            self.connection.commit()
+            logger.info('TimescaleDB table <%s> created', table_name)
 
-        except psycopg2.Error as e:
+        except PSQL_Error as e:
             logger.error("{}: {}".format(type(e).__name__, e))
             raise
 
@@ -351,7 +413,7 @@ class Database:
             logger.exception("{}: {}".format(type(e).__name__, e))
             raise
 
-    def check_table(self, table_name):
+    def check_table(self, table_name) -> bool:
         """ Checks if a table exists.
 
         Parameters
@@ -361,8 +423,8 @@ class Database:
 
         Returns
         -------
-        bool
-            True if the table exists, False otherwise.
+        bool:
+            `True` if the table exists, `False` otherwise.
 
         Raises
         ------
@@ -370,28 +432,29 @@ class Database:
             Base exception for all kind of database errors
         """
 
-        logger = logging.getLogger(__name__)
+        logger = getLogger()
 
-        if self._connection is None:
-            raise psycopg2.DatabaseError('Trying to access database without initialization')
+        # Check database status
+        if self.connection is None:
+            raise DatabaseError('Trying to access database without initialization')
 
         try:
-            logger.debug('Checking if table %s exists', table_name)
-            self._cursor.execute(
+            logger.debug('Checking if table <%s> exists', table_name)
+            self.cursor.execute(
                 self._query_checkTable,
                 {'table_schema':    'public',
                  'table_name':      table_name,
                  }
             )
-            reply = self._cursor.fetchone()
+            reply = self.cursor.fetchone()
             if reply[0]:
-                logger.debug('Table %s does exist', table_name)
+                logger.debug('Table <%s> does exist', table_name)
                 return True
             else:
-                logger.debug('Table %s does not exist', table_name)
+                logger.debug('Table <%s> does not exist', table_name)
                 return False
 
-        except psycopg2.Error as e:
+        except PSQL_Error as e:
             logger.error("{}: {}".format(type(e).__name__, e))
             raise
 
@@ -399,7 +462,7 @@ class Database:
             logger.exception("{}: {}".format(type(e).__name__, e))
             raise
 
-    def check_column(self, table_name, column_name):
+    def check_column(self, table_name, column_name) -> bool:
         """ Checks if a column exists in a given table.
 
         Parameters
@@ -411,8 +474,8 @@ class Database:
 
         Returns
         -------
-        bool
-            True if the column exists, False otherwise.
+        bool:
+            `True` if the column exists, `False` otherwise.
 
         Raises
         ------
@@ -420,29 +483,30 @@ class Database:
             Base exception for all kind of database errors
         """
 
-        logger = logging.getLogger(__name__)
+        logger = getLogger()
 
-        if self._connection is None:
-            raise psycopg2.DatabaseError('Trying to access a database without initialization')
+        # Check database status
+        if self.connection is None:
+            raise DatabaseError('Trying to access a database without initialization')
 
         try:
-            logger.info('Checking if column %s exists in table %s', column_name, table_name)
-            self._cursor.execute(
+            logger.debug('Checking if column <%s> exists in table <%s>', column_name, table_name)
+            self.cursor.execute(
                 self._query_checkColumn,
                 {'table_schema':    'public',
                  'table_name':      table_name,
                  'column_name':     column_name,
                  }
             )
-            reply = self._cursor.fetchone()
+            reply = self.cursor.fetchone()
             if reply[0]:
-                logger.debug('Column %s does exist in table %s', column_name, table_name)
+                logger.debug('Column <%s> does exist in table <%s>', column_name, table_name)
                 return True
             else:
-                logger.debug('Column %s does not exist in table %s', column_name, table_name)
+                logger.debug('Column <%s> does not exist in table <%s>', column_name, table_name)
                 return False
 
-        except psycopg2.Error as e:
+        except PSQL_Error as e:
             logger.error("{}: {}".format(type(e).__name__, e))
             raise
 
@@ -459,8 +523,9 @@ class Database:
     ):
         """ Creates a new column in a given table.
 
-        If the column already exists, it just returns True.
-        If the table does not exist, returns False.
+        If the column already exists, it just returns.
+        If the table does not exist or the column could not
+        be created, it raises a :class:`psycopg2.Error`.
 
         Parameters
         ----------
@@ -482,24 +547,25 @@ class Database:
         :class:`psycopg2.Error`
             Base exception for all kind of database errors
         """
-        logger = logging.getLogger(__name__)
 
-        if self._connection is None:
-            raise psycopg2.DatabaseError('Trying to access database without initialization')
+        logger = getLogger()
 
-        logger.info('Creating column %s in table %s', column_name, table_name)
+        # Check database status
+        if self.connection is None:
+            raise DatabaseError('Trying to access database without initialization')
 
-        # First let's check if the table exists
+        # Check if the table exists
         if not self.check_table(table_name):
-            raise psycopg2.IntegrityError(
-                'Trying to create a column in table %s, which does not exist'.format(table_name))
+            raise IntegrityError(
+                'Trying to create a column in table <%s>, which does not exist'.format(table_name))
 
         # Check if the column already exists
         if self.check_column(table_name, column_name):
-            logger.info('Column %s already exists in table %s', column_name, table_name)
+            logger.info('Column <%s> already exists in table <%s>', column_name, table_name)
             return
 
         # Now let's send the SQL query
+        logger.info('Creating column <%s> in table <%s>', column_name, table_name)
         try:
             if constraints is None:
                 constraint_list = sql.SQL(' ')
@@ -515,21 +581,23 @@ class Database:
                     table_name=table_name,
                     column_name=column_name,
                     data_type=str(data_type),
-                    checks=constraint_list.as_string(self._connection),
+                    checks=constraint_list.as_string(self.connection),
                 )
             )
+            logger.debug('Sending query: %s', query)
 
-            self._cursor.execute(query)
-            self._connection.commit()
+            self.cursor.execute(query)
+            self.connection.commit()
 
             # Check if the column was created
             if self.check_column(table_name, column_name):
-                logger.debug('Column %s successfully created in table %s', column_name, table_name)
+                logger.info('Column <%s> successfully created in table <%s>', column_name, table_name)
             else:
-                raise psycopg2.OperationalError(
-                    'Column {c} could not be created in table {t}'.format(c=column_name, t=table_name))
+                raise OperationalError(
+                    'Column <%s> could not be created in table <%s>', column_name, table_name
+                )
 
-        except (psycopg2.Error, TypeError, ValueError, KeyError) as e:
+        except (PSQL_Error, TypeError, ValueError, KeyError) as e:
             logger.error("{}: {}".format(type(e).__name__, e))
             raise
 
@@ -570,27 +638,29 @@ class Database:
         :class:`psycopg2.Error`
             Base exception for all kind of database errors
         """
-        logger = logging.getLogger(__name__)
 
-        if self._connection is None:
-            raise psycopg2.DatabaseError('Trying to access database without initialization')
+        logger = getLogger()
 
-        logger.debug('Inserting data into table %s', table_name)
+        if self.connection is None:
+            raise DatabaseError('Trying to access database without initialization')
+
+        # Check the list sizes are the same
+        if len(data) != len(columns):
+            raise DataError('Data and column lists have different sizes')
+
+        logger.debug('Inserting data into table <%s>: ', table_name)
+        for c, d in zip(columns, data):
+            logger.debug('\t{:<25}{}'.format(c, d))
 
         # Check if the table and the columns exist
         if check_columns:
             if not self.check_table(table_name):
-                raise psycopg2.IntegrityError(
-                    'Table %s does not exist'.format(table_name))
+                raise IntegrityError(
+                    'Table {} does not exist'.format(table_name))
 
             for column in columns:
                 if not self.check_column(table_name, column):
-                    raise psycopg2.IntegrityError(
-                        'Column %s does not exist in table %s'.format(column, table_name))
-
-        # Check the list sizes are the same
-        if len(data) != len(columns):
-            raise psycopg2.DataError('Data and column list have different sizes')
+                    raise IntegrityError('Column <%s> does not exist in table <%s>', column, table_name)
 
         # Now let's send the SQL query
         try:
@@ -604,15 +674,17 @@ class Database:
 
             query = sql.SQL(self._query_insertData.format(
                     table_name=table_name,
-                    columns=column_list.as_string(self._connection),
-                    data=values_list.as_string(self._connection),
+                    columns=column_list.as_string(self.connection),
+                    data=values_list.as_string(self.connection),
                 )
             )
+            logger.debug('Executing query: %s', query)
+            logger.debug('with data: %s', data)
 
-            self._cursor.execute(query, data)
-            self._connection.commit()
+            self.cursor.execute(query, data)
+            self.connection.commit()
 
-        except (psycopg2.Error, TypeError, ValueError) as e:
+        except (PSQL_Error, TypeError, ValueError) as e:
             logger.error("{}: {}".format(type(e).__name__, e))
             raise
 
@@ -620,3 +692,5 @@ class Database:
             logger.exception("{}: {}".format(type(e).__name__, e))
             raise
 
+        else:
+            logger.debug('Data successfully committed')
