@@ -31,14 +31,21 @@ buffer_size: int, 1024
     Maximum length of a transmitted messages
 """
 
+# Imports
+from re import compile
+from typing import Pattern
 import signal
 import socket
 import select
 import configparser
-import logging
-from pkg_resources import resource_filename
-from zc.lockfile import LockFile, LockError
+from os.path import abspath, expanduser
 import argparse
+
+# Third party
+from zc.lockfile import LockFile, LockError
+
+# Local packages
+from lab_utils.custom_logging import getLogger, CustomLogger
 from .__project__ import __documentation__ as docs_url
 
 # Maximum length of transmitted messages
@@ -175,8 +182,6 @@ class Server:
     The help string by the :obj:`argparse` library
     is not complete and hence is overridden by the
     :meth:`ArgumentParser.full_help` method.
-
-    TODO: Create some examples of message parsing.
     """
 
     # Flags
@@ -188,26 +193,31 @@ class Server:
     sock: socket.SocketType = None      #: Connection socket.
     address: str = None                 #: TCP binding address.
     max_backlog: int = 1                #: TCP connection queue.
-    socket_timeout: float = 1.         #: Socket time-out, used for Ctrl+C handling
+    socket_timeout: float = 1.          #: Socket time-out, used for Ctrl+C handling
 
     # Server variables
-    logger: logging.Logger = None       #: Single logger for the whole class.
+    logger: CustomLogger = None         #: Single :class:`~lab_utils.custom_logging.CustomLogger` for the whole class.
     message: str = ''                   #: Message from the client.
     reply: str = ''                     #: Reply to the client.
 
     # Daemon
     pid_file_name: str = '/tmp/socket_comm.pid'     #: The PID file name
-    lock: LockFile = None                           #: LockFile object.
+    lock: LockFile = None                           #: :class:`~zc.lockfile.LockFile` object.
 
     # Message parsing
     namespace: argparse.Namespace = None    #: Container to hold message options.
     parser: ArgumentParser = None           #: Argument parser.
+    delimiter_left: chr = None              #: Left delimiter
+    delimiter_right: chr = None             #: Left delimiter
+    regex: Pattern = None                   #: Parsing pattern.
     # noinspection PyProtectedMember
     sp: argparse._SubParsersAction = None   #: Argument subparser
 
     def __init__(self,
                  config_file: str = None,
-                 pid_file_name: str = None):
+                 pid_file_name: str = None,
+                 host: str = None,
+                 port: int = None):
         """ Initializes and runs the :class:`Server` object.
         The constructor calls the :meth:`config` method to
         read out the server attributes, and initializes
@@ -219,9 +229,16 @@ class Server:
         ----------
         config_file : str, optional
             Configuration file, default is `None`.
+
         pid_file_name : str, optional
             If given, overrides the default :attr:`PID file name<pid_file_name>`.
 
+        host : int, optional
+            If given, overrides the server :attr:`host`.
+
+        port : int, optional
+            If given, overrides the server :attr:`port`.
+
         Raises
         ------
         :class:`configparser.Error`
@@ -232,6 +249,9 @@ class Server:
             Various socket errors, e.g. address or timeout
         """
 
+        # Use a single logger for all server messages
+        self.logger = getLogger()
+
         # Read _config file
         if config_file is not None:
             self.config(config_file)
@@ -240,11 +260,21 @@ class Server:
         if pid_file_name is not None:
             self.pid_file_name = pid_file_name
 
-        # Use a single logger for all server messages
-        self.logger = logging.getLogger('Server on {h}:{p}'.format(h=self.host, p=self.port))
+        # Override server address
+        if host is not None:
+            self.host = host
+
+        if port is not None:
+            self.port = port
 
         # Initialize parser
         self.create_parser()
+        self.delimiter_left = '{'
+        self.delimiter_right = '}'
+        self.regex = compile(" (?![^{dl}{dr}]*{dr})".format(
+            dl=self.delimiter_left,
+            dr=self.delimiter_right
+        ))
 
         # Lock the PID file, raise LockError if it fails
         self.daemonize()
@@ -252,24 +282,6 @@ class Server:
         # Event flags
         self.quit_flag = False
 
-    def __del__(self):
-        """ Releases the PID lock file
-        and the TCP socket. """
-
-        if self.address is not None:
-            try:
-                self.logger.info('Closing socket')
-                self.sock.close()
-            except OSError:
-                pass
-
-        if self.lock is not None:
-            try:
-                self.logger.info('Releasing PID file')
-                self.lock.close()
-            except ValueError:
-                pass
-
     def daemonize(self):
         """ Locks a PID file to ensure that a
         single instance of the server is running.
@@ -307,9 +319,11 @@ class Server:
             If an error happened while parsing the file, e.g. no file was found
         """
 
+        # Expand configuration file path
+        filename = abspath(expanduser(filename))
+
         # Use a logger named like the module itself
-        logger = logging.getLogger(__name__)
-        logger.info("Loading configuration file %s", filename)
+        self.logger.info("Loading configuration file %s", filename)
 
         try:
             # Initialize _config parser and read file
@@ -321,12 +335,12 @@ class Server:
             self.port = config_parser.getint(section='Overall', option='port', fallback=1507)
 
         except configparser.Error as e:
-            logger.error("{}: {}".format(type(e).__name__, e))
+            self.logger.error("{}: {}".format(type(e).__name__, e))
             raise
 
         except BaseException as e:
             # Undefined exception, full traceback to be printed
-            logger.exception("{}: {}".format(type(e).__name__, e))
+            self.logger.exception("{}: {}".format(type(e).__name__, e))
             raise
 
     def start_daemon(self):
@@ -376,7 +390,6 @@ class Server:
 
                 # Check for Ctrl+C
                 if self.quit_flag:
-                    print('quit quit!')
                     break
 
                 # Connection ready?
@@ -389,12 +402,17 @@ class Server:
                     connection, client_address = self.sock.accept()
                     self.logger.debug('Client connected from %s', client_address)
                     self.message = connection.recv(buffer_size).decode().rstrip()
-                    self.logger.debug('Message: %s', self.message.split())
+
+                    # Split message and remove delimiters
+                    parsed_message = [item.strip(' {}{}'.format(self.delimiter_left, self.delimiter_right))
+                                      for item in self.regex.split(self.message)]
+                    self.logger.debug('Message: %s', parsed_message)
 
                     # Parse message and call appropriate task
                     try:
+                        self.reply = ''
                         self.namespace = self.parser.parse_args(
-                            args=self.message.split(),
+                            args=parsed_message,
                         )
                         self.namespace.func()
 
@@ -449,7 +467,26 @@ class Server:
             self.logger.exception("{}: {}".format(type(e).__name__, e))
             raise
 
-        # Server's End-Of-Life
+        finally:
+            # Server's End-Of-Life
+            self.close()
+
+    def close(self):
+        """ Releases the PID lock file
+        and the TCP socket. """
+        if self.address is not None:
+            try:
+                self.logger.info('Closing socket')
+                self.sock.close()
+            except OSError as e:
+                self.logger.error('Something went wrong... {}: {}'.format(type(e).__name__, e))
+
+        if self.lock is not None:
+            try:
+                self.logger.info('Releasing PID file')
+                self.lock.close()
+            except ValueError as e:
+                self.logger.error('Something went wrong... {}: {}'.format(type(e).__name__, e))
 
     def create_parser(self):
         """ Configures the message :attr:`parser`,
@@ -516,7 +553,12 @@ class Client:
     host: str = 'localhost'     #: Host address.
     port: int = 1507            #: Connection port.
 
-    def __init__(self, config_file: str = None, host: str = None, port: int = None):
+    def __init__(
+            self,
+            config_file: str = None,
+            host: str = None,
+            port: int = None
+    ):
         """ Initializes the :class:`Client` object. If a
         :paramref:`~Client.__init__.config_file`
         is given, the constructor calls the
@@ -546,7 +588,7 @@ class Client:
             :meth:`config` can fail raising this exception.
         """
 
-        # Read _config file, if given
+        # Read configuration file
         if config_file is not None:
             self.config(config_file)
 
@@ -557,18 +599,17 @@ class Client:
         if port is not None:
             self.port = port
 
-    def config(self, config_file: str = resource_filename(__name__, 'conf/server.ini')):
-        """ Loads the configuration from a file.
+    def config(self, config_file: str):
+        """ Loads the TCP configuration from the file
+        :paramref:`~Client.config.config_file`.
 
-        The method reads the :paramref:`~Client.config.config_file`
-        using the library :obj:`configparser`. The
-        structure of the file is shown in the
-        :ref:`examples section<configuration-files>`.
+        The method reads the file using the library
+        :obj:`configparser`..
 
         Parameters
         ----------
-        config_file : str, optional
-            TODO
+        config_file : str
+            TCP configuration file, full path.
 
         Raises
         ------
@@ -577,9 +618,9 @@ class Client:
             a parameter is missing or it has an invalid value.
         """
 
-        # Use a logger named like the module itself
-        logger = logging.getLogger(__name__)
-        logger.info("Loading configuration file %s", config_file)
+        # Expand configuration file path
+        config_file = abspath(expanduser(config_file))
+        getLogger().info("Loading configuration file %s", config_file)
 
         try:
             # Initialize _config parser and read file
@@ -591,16 +632,16 @@ class Client:
             self.port = config_parser.getint(section='Overall', option='port')
 
         except configparser.Error as e:
-            logger.error("{}: {}".format(type(e).__name__, e))
+            getLogger().error("{}: {}".format(type(e).__name__, e))
             raise
 
         except BaseException as e:
             # Undefined exception, full traceback to be printed
-            logger.exception("{}: {}".format(type(e).__name__, e))
+            getLogger().exception("{}: {}".format(type(e).__name__, e))
             raise
 
         else:
-            logger.info("Configuration file loaded")
+            getLogger().info("Configuration file loaded")
 
     def send_message(self, message: str) -> str:
         """ Complete communication process. Connects
@@ -625,8 +666,7 @@ class Client:
         """
 
         # Get the logger
-        logger = logging.getLogger(__name__)
-        logger.info('Sending message to the server: %s', message)
+        getLogger().info('Sending message to the server: %s', message)
 
         try:
             # Connect to the server
@@ -638,18 +678,18 @@ class Client:
 
             # Get reply
             reply = sock.recv(buffer_size).decode()
-            logger.info('Reply received: %s', reply)
+            getLogger().info('Reply received: %s', reply)
 
             # Close the connection
             sock.close()
 
         except OSError as e:
-            logger.error("{}: {}".format(type(e).__name__, e))
+            getLogger().error("{}: {}".format(type(e).__name__, e))
             raise
 
         except BaseException as e:
             # Undefined exception, full traceback to be printed
-            logger.exception("{}: {}".format(type(e).__name__, e))
+            getLogger().exception("{}: {}".format(type(e).__name__, e))
             raise
 
         else:
